# feature_extractor.py
import pandas as pd
from nkululeko.util import Util 
from nkululeko.feats_opensmile import Opensmileset


class FeatureExtractor:
    df = None # pandas dataframe to store the features (and indexed with the data from the sets)
    data_df = None # dataframe to get audio paths

# def __init__
    def set_data(self, data_df, data_name, feats_designation):
        self.data_df = data_df
        self.data_name = data_name
        self.util = Util()
        self.feats_designation = feats_designation

    def extract(self):
        strategy = self.util.config_val('DATA', 'strategy', 'traintest')
        feats_types = self.util.config_val_list('FEATS', 'type', ['os'])
        self.featExtractor = None
        self.feats= pd.DataFrame()
        _scale = True
        for feats_type in feats_types:
            store_name = f'{self.data_name}_{strategy}_{feats_type}'   
            if feats_type=='os':
                self.featExtractor = Opensmileset(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='trill':
                from nkululeko.feats_trill import TRILLset
                self.featExtractor = TRILLset(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='wav2vec':
                from nkululeko.feats_wav2vec2 import Wav2vec2
                self.featExtractor = Wav2vec2(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='audmodel':
                from nkululeko.feats_audmodel import AudModelSet
                self.featExtractor = AudModelSet(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='praat':
                from nkululeko.feats_praat import Praatset
                self.featExtractor = Praatset(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='mld':
                from nkululeko.feats_mld import MLD_set
                self.featExtractor = MLD_set(f'{store_name}_{self.feats_designation}', self.data_df)
            elif feats_type=='import':
                from nkululeko.feats_import import Importset
                self.featExtractor = Importset(f'{store_name}_{self.feats_designation}', self.data_df)
            else:
                self.util.error(f'unknown feats_type: {feats_type}')

            self.featExtractor.extract()
            self.featExtractor.filter()
            # remove samples that were not extracted by MLD
            #self.df_test = self.df_test.loc[self.df_test.index.intersection(featExtractor_test.df.index)]
            #self.df_train = self.df_train.loc[self.df_train.index.intersection(featExtractor_train.df.index)]
            self.util.debug(f'{feats_type}: shape : {self.featExtractor.df.shape}')
            self.feats = pd.concat([self.feats, self.featExtractor.df], axis = 1)
        return self.feats

    def extract_sample(self, signal, sr):
        return self.featExtractor.extract_sample(signal, sr)


# feats_praat.py
from nkululeko.featureset import Featureset
import os
import pandas as pd
import nkululeko.glob_conf as glob_conf
from nkululeko import feinberg_praat

class Praatset(Featureset):
    """
    a feature extractor for the Praat software, based on 
    David R. Feinberg's Praat scripts for the parselmouth python interface.
    https://osf.io/6dwr3/

    """
    def __init__(self, name, data_df):
        super().__init__(name, data_df)

    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path('store')
        storage = f'{store}{self.name}.pkl'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug('extracting Praat features, this might take a while...')
            self.df = feinberg_praat.compute_features(self.data_df.index)
            self.df = self.df.set_index(self.data_df.index)
            self.df.to_pickle(storage)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug('reusing extracted Praat features.')
            self.df = pd.read_pickle(storage)
        self.util.debug(f'praat feature names: {self.df.columns}')
        self.df = self.df.astype(float)



    def extract_sample(self, signal, sr):
        self.util.error('feats_praat: extracting single samples not implemented yet')
        feats = None
        return feats


# opensmileset.py
from nkululeko.featureset import Featureset
import opensmile
import os
import pandas as pd
import nkululeko.glob_conf as glob_conf


class Opensmileset(Featureset):

    def __init__(self, name, data_df):
        super().__init__(name, data_df)
        self.featset = self.util.config_val('FEATS', 'set', 'eGeMAPSv02')
        try:
            self.feature_set = eval(f'opensmile.FeatureSet.{self.featset}')
            #'eGeMAPSv02, ComParE_2016, GeMAPSv01a, eGeMAPSv01a':
        except AttributeError:        
            self.util.error(f'something is wrong with feature set: {self.featset}')        
        self.featlevel = self.util.config_val('FEATS', 'level', 'functionals')
        try:
            self.featlevel = self.featlevel.replace('lld', 'LowLevelDescriptors')
            self.featlevel = self.featlevel.replace('functionals', 'Functionals')
            self.feature_level = eval(f'opensmile.FeatureLevel.{self.featlevel}')
        except AttributeError:        
            self.util.error(f'something is wrong with feature level: {self.featlevel}')        


    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path('store')
        storage = f'{store}{self.name}_{self.featset}.pkl'
        extract = eval(self.util.config_val('FEATS', 'needs_feature_extraction', 'False'))
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or not os.path.isfile(storage) or no_reuse:
            self.util.debug('extracting openSmile features, this might take a while...')
            smile = opensmile.Smile(
            feature_set= self.feature_set,
            feature_level=self.feature_level,
            num_workers=5,)
            if isinstance(self.data_df.index, pd.MultiIndex):
                self.df = smile.process_index(self.data_df.index)
                self.df = self.df.set_index(self.data_df.index)
            else:
                self.df = smile.process_files(self.data_df.index)
                self.df.index = self.df.index.droplevel(1)
                self.df.index = self.df.index.droplevel(1)
            self.df.to_pickle(storage)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'False'
            except KeyError:
                pass
        else:
            self.util.debug('reusing extracted OS features.')
            self.df = pd.read_pickle(storage)


    def extract_sample(self, signal, sr):
        smile = opensmile.Smile(
                feature_set=self.feature_set,
                feature_level=opensmile.FeatureLevel.Functionals,)
        feats = smile.process_signal(signal, sr)
        return feats.to_numpy()


import audformat
import pandas as pd
import nkululeko.glob_conf as glob_conf

def limit_speakers(df, max=20):
    """ limit number of samples per speaker
        the samples are selected randomly          
    """
    df_ret = pd.DataFrame()
    for s in df.speaker.unique():
        s_df = df[df['speaker'].eq(s)]
        if s_df.shape[0] < max:
            df_ret = df_ret.append(s_df)
        else:
            df_ret = df_ret.append(s_df.sample(max))
    return df_ret

def filter_min_dur(df, min_dur):
    """remove all samples less than min_dur duration
    """
    df_ret = df.copy()
    if not isinstance(df.index, pd.MultiIndex):
        glob_conf.util.debug('converting file index to multi index, this might take a while...')
        df_ret.index = audformat.utils.to_segmented_index(df.index, allow_nat=False)    
    for i in df_ret.index:
        start = i[1]
        end = i[2]
        dur = (end - start).total_seconds()
        if dur < float(min_dur):
            df_ret = df_ret.drop(i, axis=0)
    df_ret.is_labeled = df.is_labeled
    df_ret.got_gender = df.got_gender
    df_ret.got_speaker = df.got_speaker
    return df_ret

def filter_max_dur(df, max_dur):
    """remove all samples less than min_dur duration
    """
    df_ret = df.copy()
    if not isinstance(df.index, pd.MultiIndex):
        glob_conf.util.debug('converting file index to multi index, this might take a while...')
        df_ret.index = audformat.utils.to_segmented_index(df.index, allow_nat=False)    
    for i in df_ret.index:
        start = i[1]
        end = i[2]
        dur = (end - start).total_seconds()
        if dur > float(max_dur):
            df_ret = df_ret.drop(i, axis=0)
    df_ret.is_labeled = df.is_labeled
    df_ret.got_gender = df.got_gender
    df_ret.got_speaker = df.got_speaker
    return df_ret



# model.py
from nkululeko.util import Util 
import pandas as pd
import numpy as np
import nkululeko.glob_conf as glob_conf
import sklearn.utils
from nkululeko.reporter import Reporter
import ast
from sklearn.model_selection import GridSearchCV
import pickle
import random
from sklearn.model_selection import LeaveOneGroupOut
from sklearn.model_selection import StratifiedKFold


class Model:
    """Generic model class for linear (non-neural) algorithms"""

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes"""
        self.df_train, self.df_test, self.feats_train, self.feats_test = df_train, df_test, feats_train, feats_test
        self.util = Util()
        self.target = self.util.config_val('DATA', 'target', 'emotion')
        self.run = 0
        self.epoch = 0
        self.loso = self.util.config_val('MODEL', 'loso', False)
        self.logo = self.util.config_val('MODEL', 'logo', False)
        self.xfoldx = self.util.config_val('MODEL', 'k_fold_cross', False)
                
    def set_testdata(self, data_df, feats_df):
        self.df_test, self.feats_test = data_df, feats_df

    def reset_test(self,  df_test, feats_test):
        self.df_test, self.feats_test = df_test, feats_test


    def set_id(self, run, epoch):
        self.run = run
        self.epoch = epoch
        dir = self.util.get_path('model_dir')
        name = f'{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model'
        self.store_path = dir+name


    def _x_fold_cross(self):
        # ignore train and test sets and do a "leave one speaker out"  evaluation
        self.util.debug(f'ignoring splits and doing {self.xfoldx} fold cross validation')
        feats = self.feats_train.append(self.feats_test)
        annos = self.df_train.append(self.df_test)
        targets = annos[self.target]
        _skf = StratifiedKFold(n_splits=int(self.xfoldx))
        truths, preds, results = [], [], []
        g_index = 0
        # leave-one-speaker loop    
        for train_index, test_index in _skf.split(
            feats, 
            targets,
        ):
            train_x = feats.iloc[train_index].to_numpy()
            train_y = targets[train_index]
            self.clf.fit(train_x, train_y)
            truth_x = feats.iloc[test_index].to_numpy()
            truth_y = targets[test_index]
            predict_y = self.clf.predict(truth_x)
            report = Reporter(truth_y.astype(float), predict_y, self.run, self.epoch)
            self.util.debug(f'result for fold {g_index}: {report.get_result().get_test_result()} ')
            results.append(float(report.get_result().test))            
            truths.append(truth_y)
            preds.append(predict_y)
            g_index += 1
            
        # combine speaker folds
        truth = pd.concat(truths)
        truth.name = 'truth'
        pred = pd.Series(
            np.concatenate(preds),
            index=truth.index,
            name='prediction',
        )
        self.truths = truth
        self.preds = pred        
        results = np.asarray(results)
        self.util.debug(f'KFOLD: {self.xfoldx} folds: mean {results.mean():.3f}, std: {results.std():.3f}')

    def _loso(self):
        # ignore train and test sets and do a "leave one speaker out"  evaluation
        self.util.debug('ignoring splits and doing LOSO')
        feats = self.feats_train.append(self.feats_test)
        annos = self.df_train.append(self.df_test)
        targets = annos[self.target]
        _logo = LeaveOneGroupOut()
        truths, preds, results = [], [], []
        speakers = annos['speaker']
        g_index = 0
        # leave-one-speaker loop    
        for train_index, test_index in _logo.split(
            feats, 
            targets, 
            groups=speakers,
        ):
            train_x = feats.iloc[train_index].to_numpy()
            train_y = targets[train_index]
            self.clf.fit(train_x, train_y)
            
            truth_x = feats.iloc[test_index].to_numpy()
            truth_y = targets[test_index]
            predict_y = self.clf.predict(truth_x)
            report = Reporter(truth_y.astype(float), predict_y, self.run, self.epoch)
            self.util.debug(f'result for speaker {g_index}: {report.get_result().get_test_result()} ')
            truths.append(truth_y)
            preds.append(predict_y)
            g_index += 1
            results.append(float(report.get_result().test))            
        # combine speaker folds
        truth = pd.concat(truths)
        truth.name = 'truth'
        pred = pd.Series(
            np.concatenate(preds),
            index=truth.index,
            name='prediction',
        )
        self.truths = truth
        self.preds = pred        
        results = np.asarray(results)
        self.util.debug(f'LOSO: {self.loso} folds: mean {results.mean():.3f}, std: {results.std():.3f}')

    def _do_logo(self):
        # ignore train and test sets and do a "leave one speaker group out"  evaluation
        self.util.debug('ignoring splits and doing LOSGO')
        feats = self.feats_train.append(self.feats_test)
        annos = self.df_train.append(self.df_test)
        targets = annos[self.target]
        _logo = LeaveOneGroupOut()
        truths, preds, results = [], [], []
        # get unique list of speakers
        speakers = annos['speaker'].unique()
        # create a random dictionary of groups
        sdict = {}
        for i, s in enumerate(speakers):
            sdict[s] = random.sample(range(int(self.logo)), 1)[0]    
        # add this to the annotations  
        annos['speaker_groups'] = annos['speaker'].apply(lambda x: str(sdict[x]))
        speaker_groups = annos['speaker_groups']
        g_index = 0
        # leave-one-speaker loop    
        for train_index, test_index in _logo.split(
            feats, 
            targets, 
            groups=speaker_groups,
        ):
            train_x = feats.iloc[train_index].to_numpy()
            train_y = targets[train_index]
            self.clf.fit(train_x, train_y)
            
            truth_x = feats.iloc[test_index].to_numpy()
            truth_y = targets[test_index]
            predict_y = self.clf.predict(truth_x)
            report = Reporter(truth_y.astype(float), predict_y, self.run, self.epoch)
            result = report.get_result().get_test_result()
            self.util.debug(f'result for speaker group {g_index}: {result} ')
            results.append(float(report.get_result().test))
            truths.append(truth_y)
            preds.append(predict_y)
            g_index += 1
            
        # combine speaker folds
        truth = pd.concat(truths)
        truth.name = 'truth'
        pred = pd.Series(
            np.concatenate(preds),
            index=truth.index,
            name='prediction',
        )
        self.truths = truth
        self.preds = pred
        results = np.asarray(results)
        self.util.debug(f'LOGO: {self.logo} folds: mean {results.mean():.3f}, std: {results.std():.3f}')

    def train(self):
        """Train the model"""
        # # first check if the model already has been trained
        # if os.path.isfile(self.store_path):
        #     self.load(self.run, self.epoch)
        #     self.util.debug(f'reusing model: {self.store_path}')
        #     return

        # first check if leave on  speaker out is wanted
        if self.loso:
            self._loso()
            return
        # then if leave one speaker group out validation is wanted
        if self.logo:
            self._do_logo()
            return
        # then if x fold cross validation is wanted
        if self.xfoldx:
            self._x_fold_cross()
            return

        # check for NANs in the features
        # set up the data_loaders
        if self.feats_train.isna().to_numpy().any():
            self.util.debug(f'Model, train: replacing {self.feats_train.isna().sum().sum()} NANs with 0')
            self.feats_train = self.feats_train.fillna(0)
        # remove labels from features
        feats = self.feats_train.to_numpy()
        # compute class weights
        if self.util.config_val('MODEL', 'class_weight', False):      
            self.classes_weights = sklearn.utils.class_weight.compute_sample_weight(
                class_weight='balanced',
                y=self.df_train[self.target]
            )

        tuning_params = self.util.config_val('MODEL', 'tuning_params', False)
        if tuning_params:
            # tune the model meta parameters
            tuning_params = ast.literal_eval(tuning_params)
            tuned_params={}
            try:
                scoring = glob_conf.config['MODEL']['scoring']
            except KeyError:
                self.util.error('got tuning params but no scoring')
            for param in tuning_params:
                values = ast.literal_eval(glob_conf.config['MODEL'][param])
                tuned_params[param] = values
            self.util.debug(f'tuning on {tuned_params}')
            self.clf = GridSearchCV(self.clf, tuned_params, refit = True, verbose = 3, scoring=scoring)
            try: 
                class_weight = self.util.config_val('MODEL', 'class_weight', False)
                if class_weight:
                    self.util.debug('using class weight')
                    self.clf.fit(feats, self.df_train[self.target], sample_weight=self.classes_weights)
                else:
                    self.clf.fit(feats, self.df_train[self.target])    
            except KeyError:
                self.clf.fit(feats, self.df_train[self.target])
            self.util.debug(f'winner parameters: {self.clf.best_params_}')
        else:
            class_weight = self.util.config_val('MODEL', 'class_weight', False)
            if class_weight:
                self.util.debug('using class weight')
                self.clf.fit(feats, self.df_train[self.target], sample_weight=self.classes_weights)
            else:
                labels = self.df_train[self.target]
                self.clf.fit(feats, labels)

    def get_predictions(self):
        predictions =  self.clf.predict(self.feats_test.to_numpy())
        return predictions

    def predict(self):
        if self.feats_test.isna().to_numpy().any():
            self.util.debug(f'Model, test: replacing {self.feats_test.isna().sum().sum()} NANs with 0')
            self.feats_test = self.feats_test.fillna(0)
        if self.loso or self.logo or self.xfoldx:
            report = Reporter(self.truths.astype(float), self.preds, self.run, self.epoch)
            return report
        """Predict the whole eval feature set"""
        predictions = self.get_predictions()
        report = Reporter(self.df_test[self.target]\
            .to_numpy().astype(float), predictions, self.run, self.epoch)
        return report

    def predict_sample(self, features):
        """Predict one sample"""
        prediction = {}
        if self.util.exp_is_classification():
            # get the class probabilities
            predictions = self.clf.predict_proba(features)
            # pred = self.clf.predict(features)
            for i in range(len(self.clf.classes_)):
                cat = self.clf.classes_[i]
                prediction[cat] = predictions[0][i]
        else:
            predictions = self.clf.predict(features)
            prediction['result'] = predictions[0]
        return prediction


    def store(self):
        with open(self.store_path, 'wb') as handle:
            pickle.dump(self.clf, handle)

        
    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path('model_dir')
        name = f'{self.util.get_exp_name(only_train=True)}_{self.run}_{self.epoch:03d}.model'
        with open(dir+name, 'rb') as handle:
            self.clf = pickle.load(handle)

    def load_path(self, path, run, epoch):
        self.set_id(run, epoch)
        with open(path, 'rb') as handle:
            self.clf = pickle.load(handle)


# runmanager.py

from nkululeko.reporter import Reporter
from nkululeko.util import Util  
import nkululeko.glob_conf as glob_conf
from nkululeko.modelrunner import Modelrunner

class Runmanager:
    """Class to manage the runs of the experiment (e.g. when results differ caused by random initialization)"""
    model = None  # The underlying model
    df_train, df_test, feats_train, feats_test = None, None, None, None # The dataframes
    reports = []


    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor setting up the dataframes
        Args:
            df_train: train dataframe
            df_test: test dataframe
            feats_train: train features
            feats_train: test features

        Returns:
        
        """
        self.df_train, self.df_test, self.feats_train, self.feats_test = df_train, df_test, feats_train, feats_test
        self.util = Util()
        self.target = glob_conf.config['DATA']['target']
        # intialize a new model
        #model_type = glob_conf.config['MODEL']['type']
        #self._select_model(model_type)

        

    def do_runs(self):
        """Start the runs"""
        self.best_results = [] # keep the best result per run 
        # for all runs
        for run in range(int(self.util.config_val('EXP', 'runs', 1))):
            self.util.debug(f'run {run}')
            # set the run index as global variable for reporting
            self.util.set_config_val('EXP', 'run', run)
            self.modelrunner = Modelrunner(self.df_train, self.df_test, self.feats_train, self.feats_test, run)
            self.reports = self.modelrunner.do_epochs()
            # wrap up the run 
            plot_anim_progression = self.util.config_val('PLOT', 'anim_progression', 0)
            if plot_anim_progression:
                plot_name_suggest = self.util.get_exp_name()
                plot_name = self.util.config_val('PLOT', 'name', plot_name_suggest)+'_conf_anim.gif'
                self.util.debug(f'plotting animated confusion to {plot_name}')
                self.reports[-1].make_conf_animation(plot_name)
            plot_epoch_progression = self.util.config_val('PLOT', 'epoch_progression', 0)
            if plot_epoch_progression:
                plot_name_suggest = self.util.get_exp_name()
                plot_name = self.util.config_val('PLOT', 'name', plot_name_suggest)+'_epoch_progression'
                self.util.debug(f'plotting progression to {plot_name}')
                self.reports[-1].plot_epoch_progression(self.reports, plot_name)
            # remember the best run
            best_report = self.get_best_result(self.reports)
            plot_best_model = self.util.config_val('PLOT', 'best_model', False)
            if plot_best_model:
                plot_name_suggest = self.util.get_exp_name()
                plot_name = self.util.config_val('PLOT', 'name', plot_name_suggest)+f'_BEST_{best_report.run}_{best_report.epoch:03d}_BEST_cnf'
                self.util.debug(f'best result with run {best_report.run} and epoch {best_report.epoch}: {best_report.result.get_test_result()}')
                self.print_model(best_report, plot_name)
            # finally, print out the numbers for this run
            self.reports[-1].print_results(int(self.util.config_val('EXP', 'epochs', 1)))
            self.best_results.append(best_report)

    def print_best_result_runs(self):
        """Print the best result for all runs"""
        best_report = self.get_best_result(self.best_results)
        self.util.debug(f'best result all runs with run {best_report.run} \
            and epoch {best_report.epoch}: {best_report.result.test:.3f}')
        plot_name_suggest = self.util.get_exp_name()
        plot_name = self.util.config_val('PLOT', 'name', plot_name_suggest)+f'_BEST_{best_report.run}_{best_report.epoch:03d}_BEST_cnf'
        self.print_model(best_report, plot_name)

    def print_given_result(self, run, epoch):
        """Print a result (confusion matrix) for a given epoch and run
        Args: 
            run: for which run
            epoch: for which epoch
        
        """
        report =  Reporter([], [])
        report.set_id(run, epoch)
        self.util.debug(f'Re-testing result with run {run} and epoch {epoch}')
        plot_name_suggest = self.util.get_exp_name()
        plot_name = self.util.config_val('PLOT', 'name', plot_name_suggest)+f'_extra_{run}_{epoch:03d}_cnf'
        self.print_model(report, plot_name)

    def print_model(self, report, plot_name):
        """Print a confusion matrix for a special report
        Args:
            report: for which report (will be computed newly from model)
            plot_name: name of plot file
        """
        epoch = report.epoch
        # self.load_model(report)
        # report = self.model.predict()
        self.util.debug(f'plotting conf matrix to {plot_name}')
        report.plot_confmatrix(plot_name, epoch)
        report.print_results(epoch)


    def load_model(self, report):
        """Load a model from disk for a specific run and epoch and evaluate
        Args:
            report: for which report (will be re-evaluated)
        
        """
        run = report.run
        epoch = report.epoch
        self.util.set_config_val('EXP', 'run', run)
        model_type = glob_conf.config['MODEL']['type']
        model = self.modelrunner._select_model(model_type)
        model.load(run, epoch)
        return model

    def get_best_model(self):
        best_report = self.get_best_result(self.best_results)
        return self.load_model(best_report)

    def get_best_result(self, reports):
        best_r = Reporter([], [], 0, 0)
        if self.util.exp_is_classification():
            measure = self.util.config_val('MODEL', 'measure', 'uar')
            best_r = self.search_best_result(reports, 'ascending')
        else:
            measure = self.util.config_val('MODEL', 'measure', 'mse')
            if measure == 'mse':
                best_r = self.search_best_result(reports, 'descending')
            elif measure == 'ccc':
                best_r = self.search_best_result(reports, 'ascending')
        return best_r

    def search_best_result(self, reports, order):
        best_r = Reporter([], [], 0, 0)
        if order == 'ascending':
            best_result = 0
            for r in reports:
                res = r.result.test
                if res > best_result:
                    best_result = res
                    best_r = r
        else:
            best_result = 10000
            for r in reports:
                res = r.result.test
                if res < best_result:
                    best_result = res
                    best_r = r
        return best_r